/*
 * Copyright 2020 Masarykova univerzita
 *
 *
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package cz.muni.csirt.analyza.controller;

import cz.muni.csirt.analyza.entity.GenericObject;
import cz.muni.csirt.analyza.service.GenericObjectService;
import cz.muni.csirt.analyza.util.config.ProxyProperties;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpMethod;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import javax.servlet.http.HttpServletRequest;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collection;


/**
 * Controller for work with GenericObjects.
 */
@RestController
@RequestMapping("/generic-objects")
public class GenericObjectController {
    private ProxyProperties proxyProperties = new ProxyProperties();
    private GenericObjectService genericObjectService;
    private String warehouseLocation = proxyProperties.getLocalWarehouseUrl() + "generic-objects/";
    private RestTemplate templ = new RestTemplate();

    @Autowired
    public GenericObjectController(GenericObjectService genericObjectService) {
        this.genericObjectService = genericObjectService;
    }

    @PostMapping(produces = "application/json")
    public ResponseEntity<String> postObject(HttpMethod method,
                                            HttpServletRequest httpRequest,
                                            HttpEntity<String> httpEntity) {
        ResponseEntity<String> resp;
        String url = warehouseLocation + "?" + httpRequest.getQueryString();
        try {
            url = URLDecoder.decode(url, StandardCharsets.UTF_8.toString());
            resp = templ.exchange(url, method, httpEntity, String.class);
        } catch (UnsupportedEncodingException | RestClientException e) {
            System.err.println("Problem while resending.");
            return ResponseEntity.notFound().build();
        }
        return resp;
    }

    @ResponseBody
    @RequestMapping(value = "/{uuid}", method = RequestMethod.GET, produces = "application/json")
    public ResponseEntity<String> getObject(@PathVariable("uuid") String uuid,
                                                   HttpMethod method,
                                                   HttpEntity<GenericObject> httpEntity) {
        ResponseEntity<String> resp;
        String url = warehouseLocation + uuid;
        try{
            resp = templ.exchange(url, method, httpEntity, String.class);
        } catch (RestClientException e) {
            System.err.println("Problem while resending.");
            return ResponseEntity.notFound().build();
        }
        return resp;
    }

    @RequestMapping(value = "/{uuid}", method = RequestMethod.DELETE)
    public ResponseEntity<String> deleteObject(@PathVariable("uuid") String uuid,
                                             HttpMethod method,
                                             HttpEntity<Void> httpEntity) {
        ResponseEntity<String> resp;
        String url = warehouseLocation + uuid;
        try {
            resp = templ.exchange(url, method, httpEntity, String.class);

        } catch (RestClientException e) {
            System.err.println("Problem while resending.");
            return ResponseEntity.notFound().build();
        }
        return resp;
    }

    @ResponseBody
    @RequestMapping(value = "/filter")
    public ResponseEntity<String> filterObjects(HttpServletRequest httpRequest,
                                                                   HttpMethod method,
                                                                   HttpEntity<String> httpEntity) {
        ResponseEntity<String> resp;
        String url = warehouseLocation + "filter?" + httpRequest.getQueryString();
        try {
            url = URLDecoder.decode(url, StandardCharsets.UTF_8.toString());
            resp = templ.exchange(url, method, httpEntity, String.class);
        } catch (UnsupportedEncodingException | RestClientException e) {
            System.err.println("Problem while resending.");
            return ResponseEntity.notFound().build();
        }
        return resp;
    }

    @GetMapping(value = "/filter/{uuid}", produces = "application/json")
    public ResponseEntity<String> getObjectsByUuidFilter(@PathVariable("uuid") String uuid,
                                                              HttpMethod method,
                                                              HttpEntity<Collection<GenericObject>> httpEntity) {
        ResponseEntity<String> resp;
        String url = warehouseLocation + "filter/" + uuid;
        try {
            resp = templ.exchange(url, method, httpEntity, String.class);
        } catch (RestClientException e) {
            System.err.println("Problem while resending.");
            return ResponseEntity.notFound().build();
        }
        return resp;
    }

////////////////////////////////// Derivations

    @PostMapping(value = "/derivation", produces = "application/json")
    public ResponseEntity<String> createDerivationDetail(HttpServletRequest httpRequest,
                                                         HttpMethod method,
                                                         HttpEntity<Collection<GenericObject>> httpEntity) {
        ResponseEntity<String> resp;
        String url = warehouseLocation + "derivation?" + httpRequest.getQueryString();
        try {
            url = URLDecoder.decode(url, StandardCharsets.UTF_8.toString());
            resp = templ.exchange(url, method, httpEntity, String.class);
        } catch (UnsupportedEncodingException | RestClientException e) {
            System.err.println("Problem while resending.");
            return ResponseEntity.notFound().build();
        }
        return resp;
    }

    @GetMapping(value = "/derivation/{inputUuid}", produces = "application/json")
    public ResponseEntity<String> getDerivationDetailsByInput(@PathVariable("inputUuid") String uuid,
                                                                                 HttpMethod method,
                                                                                 HttpEntity<Collection<GenericObject>> httpEntity) {
        ResponseEntity<String> resp;
        String url = warehouseLocation + "derivation/" + uuid;
        try {
            resp = templ.exchange(url, method, httpEntity, String.class);
        } catch (RestClientException e) {
            System.err.println("Problem while resending.");
            return ResponseEntity.notFound().build();
        }
        return resp;
    }

    //////////////////////////// Hierarchy

    @ResponseBody
    @RequestMapping(value = "/filter/hierarchy/{mode}", method = RequestMethod.GET, produces = "application/json")
    public ResponseEntity<Collection<GenericObject>> getFromHierarchyByFilter(@PathVariable("mode") String mode,
                                                                            HttpServletRequest httpRequest,
                                                                            HttpEntity<String> httpEntity) {
        String specifUrl = "filter?" + httpRequest.getQueryString();
        try {
            specifUrl = URLDecoder.decode(specifUrl, StandardCharsets.UTF_8.toString());
        } catch (UnsupportedEncodingException e) {
            System.err.println("Problem with url string.");
            return ResponseEntity.notFound().build();
        }
        Collection<GenericObject> body = requestSender(specifUrl, mode, httpEntity);
        if (body == null) {
            return ResponseEntity.notFound().build();
        }
        return ResponseEntity.ok(body);
    }

    @ResponseBody
    @RequestMapping(value = "/filter/hierarchy/{mode}/{inputUuid}", method = RequestMethod.GET, produces = "application/json")
    public ResponseEntity<Collection<GenericObject>> getFromHierarchyByUuid(@PathVariable("mode") String mode,
                                                                      @PathVariable("inputUuid") String uuid,
                                                                      HttpEntity<String> httpEntity) {

        String specifUrl = "filter/" + uuid;
        Collection<GenericObject> body = requestSender(specifUrl, mode, httpEntity);
        if (body == null) {
            return ResponseEntity.notFound().build();
        }
        return ResponseEntity.ok(body);
    }

    @ResponseBody
    @RequestMapping(value = "/derivation/hierarchy/{mode}/{inputUuid}", method = RequestMethod.GET, produces = "application/json")
    public ResponseEntity<Collection<GenericObject>> getDerivationsFromHierarchyByUuid(@PathVariable("mode") String mode,
                                                             @PathVariable("inputUuid") String uuid,
                                                             HttpEntity<String> httpEntity) {
        String specifUrl = "derivation/" + uuid;
        Collection<GenericObject> body = requestSender(specifUrl, mode, httpEntity);
        if (body == null) {
            return ResponseEntity.notFound().build();
        }
        return ResponseEntity.ok(body);
    }

    private Collection<GenericObject> requestSender(String specifUrl, String mode,
                                                    HttpEntity<String> httpEntity) {
        ArrayList<String> hierarchyW = new ArrayList<>();
        ArrayList<String> hierarchyUrls = new ArrayList<>();
        String localW;
        String resp;
        String localUrl = warehouseLocation + specifUrl;
        try {
            localW = templ.exchange(localUrl, HttpMethod.GET, httpEntity, String.class).getBody();
        } catch (RestClientException e) {
            return null;
        }
        ArrayList<String> remoteWarehouses = proxyProperties.getRemoteWarehousesUrls();
        for (String warehouse : remoteWarehouses) {
            hierarchyUrls.add(warehouse + "generic-objects/" + specifUrl);
        }
        for (String url : hierarchyUrls) {
            try {
                resp = templ.exchange(url, HttpMethod.GET, httpEntity, String.class).getBody();
                hierarchyW.add(resp);
            } catch (RestClientException e) {
                hierarchyW.add("[]");
            }
        }
        return this.genericObjectService.getFullHierarchy(localW, hierarchyW, mode);
    }
}
