/*
 * Copyright 2020 Masarykova univerzita
 *
 *
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package cz.muni.csirt.analyza.controller;

import cz.muni.csirt.analyza.entity.Link;
import cz.muni.csirt.analyza.service.LinkService;
import cz.muni.csirt.analyza.util.config.ProxyProperties;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpMethod;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import javax.servlet.http.HttpServletRequest;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collection;
import java.util.UUID;


/**
 * Controller for work with Links.
 */
@RestController
@RequestMapping(value = "/links")
public class LinkController {

    private ProxyProperties proxyProperties = new ProxyProperties();
    private LinkService linkService;
    private String warehouseLocation = proxyProperties.getLocalWarehouseUrl() + "links/";
    RestTemplate templ = new RestTemplate();

    @Autowired
    public LinkController(LinkService linkService) {
        this.linkService = linkService;
    }

    @RequestMapping(method = RequestMethod.POST)
    public ResponseEntity<String> putLink(
                                        HttpMethod method,
                                        HttpEntity<String> httpEntity,
                                        HttpServletRequest httpRequest) {
        ResponseEntity<String> resp;
        String url = warehouseLocation + "?" + httpRequest.getQueryString();
        try {
            url = URLDecoder.decode(url, StandardCharsets.UTF_8.toString());
            resp = templ.exchange(url, method, httpEntity, String.class);
        } catch (UnsupportedEncodingException | RestClientException e) {
            System.err.println("Problem while resending.");
            return ResponseEntity.notFound().build();
        }
        return resp;
    }

    @ResponseBody
    @RequestMapping(value = "/{uuid}", method = RequestMethod.GET)
    public ResponseEntity<Object> getLink(@PathVariable("uuid")UUID uuid,
                                        HttpMethod method,
                                        HttpEntity<Object> httpEntity) {
        ResponseEntity<Object> resp;
        String url = warehouseLocation + uuid;

        try {
            resp = templ.exchange(url, method, httpEntity, Object.class);
        } catch (RestClientException e) {
            System.err.println("Problem while resending.");
            return ResponseEntity.notFound().build();
        }
        return resp;
    }

    @RequestMapping(value = "/{uuid}", method = RequestMethod.DELETE)
    public ResponseEntity<String> deleteLink(@PathVariable("uuid")String uuid,
                                             HttpMethod method,
                                             HttpEntity<String> httpEntity) {
        ResponseEntity<String> resp;
        String url = warehouseLocation + uuid;
        try {
            resp = templ.exchange(url, method, httpEntity, String.class);
        } catch (RestClientException e) {
            System.err.println("Problem while resending.");
            return ResponseEntity.notFound().build();
        }
        return resp;
    }

    @ResponseBody
    @RequestMapping(value = "/filter/{uuid}", method = RequestMethod.GET)
    public ResponseEntity<Collection<Object>> filterObjects(@PathVariable("uuid") String uuid,
                                                          HttpMethod method,
                                                          HttpEntity<Collection<Link>> httpEntity) {
        ResponseEntity<Collection<Object>> resp;
        String url = warehouseLocation + "filter/" + uuid;
        try {
            resp = templ.exchange(url, method, httpEntity, new ParameterizedTypeReference<Collection<Object>>() {});
        } catch (RestClientException e) {
            System.err.println("Problem while resending.");
            return ResponseEntity.notFound().build();
        }
        return resp;
    }

    /////////////////////////////////////////// Hierarchy

    @ResponseBody
    @RequestMapping(value = "/filter/hierarchy/{mode}/{uuid}", method = RequestMethod.GET, produces = "application/json")
    public ResponseEntity<Collection<Link>> getFromHierarchy(@PathVariable("mode") String mode,
                                                             @PathVariable("uuid") String uuid,
                                                             HttpMethod method,
                                                             HttpEntity<ArrayList<Link>> httpEntity) {
        RestTemplate templ = new RestTemplate();
        ArrayList<String> hierarchyW = new ArrayList<>();
        ArrayList<String> hierarchyUrls = new ArrayList<>();

        String specifUrl = "links/filter/" + uuid;
        String localUrl = warehouseLocation + "filter/" + uuid;

        try {
            String body = templ.exchange(localUrl, method, httpEntity, String.class).getBody();
            hierarchyW.add(body);
        } catch (RestClientException e) {
            return ResponseEntity.notFound().build();
        }

        ArrayList<String> remoteWarehouses = proxyProperties.getRemoteWarehousesUrls();
        for (String warehouse : remoteWarehouses) {
            hierarchyUrls.add(warehouse + specifUrl);
        }
        for (String url : hierarchyUrls) {
            try {
                String resp = templ.exchange(url, HttpMethod.GET, httpEntity, String.class).getBody();
                hierarchyW.add(resp);
            } catch (RestClientException e) {
                hierarchyW.add("[]");
            }
        }
        return ResponseEntity.ok(this.linkService.getFullHierarchy(hierarchyW, mode));
    }
}
